@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP, FlexibleInstances #-}
+{-# LANGUAGE CPP, FlexibleInstances, FlexibleContexts #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      : Data.Binary
@@ -88,6 +88,7 @@ import Data.Array.Unboxed
 --
 #if __GLASGOW_HASKELL__ >= 606
 import qualified Data.Sequence as Seq
+import qualified Data.Foldable as Fold
 #endif
 
 ------------------------------------------------------------------------
@@ -104,9 +105,10 @@ import qualified Data.Sequence as Seq
 --
 -- Instances of Binary should satisfy the following property:
 --
--- > get . put == id
+-- > decode . encode == id
 --
--- A range of instances are provided for basic Haskell types. 
+-- That is, the 'get' and 'put' methods should be the inverse of each
+-- other. A range of instances are provided for basic Haskell types. 
 --
 class Binary t where
     -- | Encode a value in the Put monad.
@@ -249,7 +251,7 @@ decode = runGet get
 encodeFile :: Binary a => FilePath -> a -> IO ()
 encodeFile f v = L.writeFile f (encode v)
 
--- | Lazily reconstruct a value previously written to a file
+-- | Lazily reconstruct a value previously written to a file.
 --
 -- This is just a convenience function, it's defined simply as:
 --
@@ -259,8 +261,20 @@ encodeFile f v = L.writeFile f (encode v)
 --
 -- > return . decode . decompress =<< B.readFile f
 --
+-- After contructing the data from the input file, 'decodeFile' checks
+-- if the file is empty, and in doing so will force the associated file
+-- handle closed, if it is indeed empty. If the file is not empty, 
+-- it is up to the decoding instance to consume the rest of the data,
+-- or otherwise finalise the resource.
+--
 decodeFile :: Binary a => FilePath -> IO a
-decodeFile f = liftM decode (L.readFile f)
+decodeFile f = do
+    s <- L.readFile f
+    return $ runGet (do v <- get
+                        m <- isEmpty
+                        m `seq` return v) s
+
+-- needs bytestring 0.9.1.x to work 
 
 ------------------------------------------------------------------------
 -- Lazy put and get
@@ -571,7 +585,18 @@ instance (Binary a, Binary b, Binary c, Binary d, Binary e,
 instance Binary a => Binary [a] where
     put l  = put (length l) >> mapM_ put l
     get    = do n <- get :: Get Int
-                replicateM n get
+                getMany n
+
+-- | 'getMany n' get 'n' elements in order, without blowing the stack.
+getMany :: Binary a => Int -> Get [a]
+getMany n = go [] n
+ where
+    go xs 0 = return $! reverse xs
+    go xs i = do x <- get
+                 -- we must seq x to avoid stack overflows due to laziness in
+                 -- (>>=)
+                 x `seq` go (x:xs) (i-1)
+{-# INLINE getMany #-}
 
 instance (Binary a) => Binary (Maybe a) where
     put Nothing  = putWord8 0
@@ -613,20 +638,20 @@ instance Binary ByteString where
 -- Maps and Sets
 
 instance (Ord a, Binary a) => Binary (Set.Set a) where
-    put = put . Set.toAscList
-    get = liftM Set.fromDistinctAscList get
+    put s = put (Set.size s) >> mapM_ put (Set.toAscList s)
+    get   = liftM Set.fromDistinctAscList get
 
 instance (Ord k, Binary k, Binary e) => Binary (Map.Map k e) where
-    put = put . Map.toAscList
-    get = liftM Map.fromDistinctAscList get
+    put m = put (Map.size m) >> mapM_ put (Map.toAscList m)
+    get   = liftM Map.fromDistinctAscList get
 
 instance Binary IntSet.IntSet where
-    put = put . IntSet.toAscList
-    get = liftM IntSet.fromDistinctAscList get
+    put s = put (IntSet.size s) >> mapM_ put (IntSet.toAscList s)
+    get   = liftM IntSet.fromDistinctAscList get
 
 instance (Binary e) => Binary (IntMap.IntMap e) where
-    put = put . IntMap.toAscList
-    get = liftM IntMap.fromDistinctAscList get
+    put m = put (IntMap.size m) >> mapM_ put (IntMap.toAscList m)
+    get   = liftM IntMap.fromDistinctAscList get
 
 ------------------------------------------------------------------------
 -- Queues and Sequences
@@ -637,12 +662,13 @@ instance (Binary e) => Binary (IntMap.IntMap e) where
 --
 
 instance (Binary e) => Binary (Seq.Seq e) where
-    -- any better way to do this?
-    put s = put . flip unfoldr s $ \sq ->
-        case Seq.viewl sq of
-            Seq.EmptyL -> Nothing
-            (Seq.:<) e sq' -> Just (e,sq')
-    get = fmap Seq.fromList get
+    put s = put (Seq.length s) >> Fold.mapM_ put s
+    get = do n <- get :: Get Int
+             rep Seq.empty n get
+      where rep xs 0 _ = return $! xs
+            rep xs n g = xs `seq` n `seq` do
+                           x <- g
+                           rep (xs Seq.|> x) (n-1) g
 
 #endif
 
@@ -675,7 +701,7 @@ instance (Binary i, Ix i, Binary e) => Binary (Array i e) where
     get = do
         bs <- get
         n  <- get                  -- read the length
-        xs <- replicateM n get     -- now the elems.
+        xs <- getMany n            -- now the elems.
         return (listArray bs xs)
 
 --
@@ -689,5 +715,5 @@ instance (Binary i, Ix i, Binary e, IArray UArray e) => Binary (UArray i e) wher
     get = do
         bs <- get
         n  <- get
-        xs <- replicateM n get
+        xs <- getMany n
         return (listArray bs xs)
